
from ..sequencer.sequencer_data import Wave, AcquisitionBins, AcquisitionWeight
from ..lang.exceptions import Q1TypeError

class GeneratorData:
    def __init__(self):
        self.waveforms = {}
        self.weights = {}
        self.acquisitions = {}

    def translate_wave(self, wave):
        if not isinstance(wave, Wave):
            raise Q1TypeError(f'Unsupported type for wave: {wave}')

        if wave.name not in self.waveforms:
            self.waveforms[wave.name] = {
                    'data':list(wave.data),
                    'index':len(self.waveforms)
                    }

        return self.waveforms[wave.name]['index']

    def translate_acquisition(self, acquisition):
        if not isinstance(acquisition, AcquisitionBins):
            raise Q1TypeError(f'Unsupported type for acquisition: {acquisition}')

        if acquisition.name not in self.acquisitions:
            self.acquisitions[acquisition.name] = {
                    'num_bins': acquisition.num_bins,
                    'index': len(self.acquisitions),
                    }

        return self.acquisitions[acquisition.name]['index']

    def translate_weight(self, weight):
        if not isinstance(weight, AcquisitionWeight):
            raise Q1TypeError(f'Unsupported type for weight: {weight}')

        if weight.name not in self.weights:
            self.weights[weight.name] = {
                    'data': list(weight.data),
                    'index': len(self.weights),
                    }

        return self.weights[weight.name]['index']

    def get_data_dict(self):
        return {
                'waveforms': self.waveforms,
                'weights': self.weights,
                'acquisitions': self.acquisitions,
                }



import os
import time
import logging
from pathlib import Path
from tempfile import TemporaryDirectory

from .program import Program
from .lang.exceptions import Q1InputOverloaded
from .sequencer.sequencer import SequenceBuilder
from .sequencer.control import ControlBuilder
from .sequencer.readout import ReadoutBuilder
from .modules.modules import QcmModule, QrmModule
from .util.qblox_version import check_qblox_instrument_version, qblox_version, Version
from qblox_instruments import InstrumentType

logger = logging.getLogger(__name__)


class Q1Instrument:
    # Postpone error checking till the end to save communication overhead.
    # System errors are only reported for SCPI errors. It's almost impossible to
    # get an error, because everything is already checked in qblox-instruments code.
    _i_feel_lucky = True

    _exception_on_overload = True

    def __init__(self, path=None, add_traceback=True):
        check_qblox_instrument_version()
        if path:
            self.path = path
        else:
            q1dir = Path.home() / '.q1'
            q1dir.mkdir(exist_ok=True)
            self.temp_dir = TemporaryDirectory(dir=q1dir)
            self.path = self.temp_dir.name
            logger.info('Instrument upload temp dir: ' + self.path)
        self.root_instruments = set()
        self.modules = {}
        self.controllers = {}
        self.readouts = {}
        self._loaded_q1asm = {}
        SequenceBuilder.add_traceback_to_instructions = add_traceback

    def add_pulsar(self, pulsar):
        if pulsar.instrument_type == InstrumentType.QCM:
            self.add_qcm(pulsar)
        elif pulsar.instrument_type == InstrumentType.QRM:
            self.add_qrm(pulsar)
        else:
            raise Exception(f'Unknown instrument type: {pulsar.instrument_type}')

    def add_qcm(self, pulsar):
        self.modules[pulsar.name] = QcmModule(pulsar)
        self._add_root_instrument(pulsar.root_instrument)

    def add_qrm(self, pulsar):
        self.modules[pulsar.name] = QrmModule(pulsar)
        self._add_root_instrument(pulsar.root_instrument)

    def add_control(self, name, module_name, channels, nco_frequency=None):
        sequencer = self.modules[module_name].get_sequencer(channels)
        sequencer.nco_frequency = nco_frequency
        self.controllers[name] = sequencer

    def add_readout(self, name, module_name, out_channels=[], nco_frequency=None):
        module = self.modules[module_name]
        if not isinstance(module, QrmModule):
            raise Exception('Module {module_name} is not a QRM')
        sequencer = module.get_sequencer(out_channels)
        sequencer.nco_frequency = nco_frequency
        self.readouts[name] = sequencer

    def _add_root_instrument(self, root_instrument):
        self.root_instruments.add(root_instrument)
        if Q1Instrument._i_feel_lucky and hasattr(root_instrument, '_debug'):
            # Change the debug level to speed up communication.
            # Errors will be checked before start of the sequence.
            root_instrument._debug = 2

    def new_program(self, prog_name):
        program = Program(path=os.path.join(self.path, prog_name))
        for name, seq in self.controllers.items():
            seq_builder = ControlBuilder(name, seq.enabled_paths,
                                         seq.max_output_voltage,
                                         seq.nco_frequency)
            program.add_sequence_builder(seq_builder)

        for name, seq in self.readouts.items():
            seq_builder = ReadoutBuilder(name, seq.enabled_paths,
                                         seq.max_output_voltage,
                                         seq.nco_frequency)
            program.add_sequence_builder(seq_builder)

        return program

    def run_program(self, program):
        self.start_program(program)
        self.wait_stopped()

    def start_program(self, program):
        t_start = time.perf_counter()

        for instrument in self.root_instruments:
            check_instrument_status(instrument)

        sequencers = { **self.controllers, **self.readouts }
        for name,seq in sequencers.items():
            module = self.modules[seq.module_name]

            q1asm = program.q1asm(name)
            self._loaded_q1asm[name] = q1asm
            if q1asm is None:
                module.disable_seq(seq)
                logger.info(f'Sequencer {name} no sequence')
                continue
            module.upload(seq.seq_nr, q1asm)
            t = (time.perf_counter() - t_start) * 1000
            logger.info(f'Sequencer {name} loaded ({t:5.3f} ms)')

            module.enable_seq(seq)
            prog_seq = program[name]
            module.set_nco(seq.seq_nr, prog_seq.nco_frequency)
            if prog_seq.modifies_frequency:
                module.invalidate_cache(seq.seq_nr, 'nco_freq')
            if prog_seq.mixer_gain_ratio is not None:
                module.set_mixer_gain_ratio(seq.seq_nr, prog_seq.mixer_gain_ratio)
            if prog_seq.mixer_phase_offset_degree is not None:
                module.set_mixer_phase_offset_degree(seq.seq_nr, prog_seq.mixer_phase_offset_degree)

        t = (time.perf_counter() - t_start) * 1000
        logger.info(f'Configure QRMs ({t:5.3f} ms)')
        for name,seq in self.readouts.items():
            readout = program[name]
            module = self.modules[seq.module_name]
            if not module.enabled(seq.seq_nr):
                continue
            module.phase_rotation_acq(seq.seq_nr, readout.phase_rotation_acq)
            module.discretization_threshold_acq(seq.seq_nr, readout.discretization_threshold_acq)
            module.integration_length_acq(seq.seq_nr, int(readout.integration_length_acq))
            if qblox_version >= Version('0.7'):
                module.delete_acquisition_data(seq.seq_nr)

        # Note: arm per sequencer. Arm on the cluster still gives red leds on the modules.
        for module in self.modules.values():
            module.arm_sequencers()
#        for instrument in self.root_instruments:
#            instrument.arm_sequencer()

        if Q1Instrument._i_feel_lucky:
            t = (time.perf_counter() - t_start) * 1000
            logger.info(f'Check status  ({t:5.3f} ms)')
            self.check_system_errors()

        for instrument in self.root_instruments:
            t = (time.perf_counter() - t_start) * 1000
            logger.info(f'Start  ({t:5.3f} ms)')
            instrument.start_sequencer()

        t = (time.perf_counter() - t_start) * 1000
        logger.info(f'Duration upload/start: ({t:5.3f}ms)')

    def wait_stopped(self, timeout_minutes=1):
        # Wait for completion
        errors = {}
        msg_level = 0
        sequencers = { **self.controllers, **self.readouts }
        for name,seq in sequencers.items():
            module = self.modules[seq.module_name]
            if not module.enabled(seq.seq_nr):
                continue
            state = module.get_sequencer_state(seq.seq_nr, timeout_minutes)
            logger.log(state.level,
                        f'Status {name} ({module.pulsar.name}:{seq.seq_nr}):'
                         f'{state}')
            msg_level = max(msg_level, state.level)
            if state.input_overloaded:
                if Q1Instrument._exception_on_overload:
                    raise Q1InputOverloaded(
                            f'INPUT OVERLOAD on {name}.'
                            '\nException can be suppressed with q1pulse.set_exception_on_overload(False)')
                else:
                    print(f'WARNING: input overload on {name}')
            if state.status != 'STOPPED' or state.level >= logging.WARNING:
                errors[name] = str(state)
        if msg_level == logging.ERROR:
            logger.error('*** Program errors ***')
            for name,state in errors.items():
                logger.error(f'  {name}: {state}')
            raise Exception(f'Q1 failures (see logger):\n {errors}')

        for instrument in self.root_instruments:
            logger.info(f'Stop')
            instrument.stop_sequencer()

    def check_system_errors(self):
        for instrument in self.root_instruments:
            errors = []
            while instrument.get_num_system_error() != 0:
                errors.append(instrument.get_system_error())

            if len(errors) > 0:
                if Q1Instrument._i_feel_lucky:
                    logger.error(f"You're not lucky. One of the previous calls failed...")
                msg = instrument.name + ':' + '\n'.join(errors)
                logger.error(msg)
                raise RuntimeError(msg)

    def get_acquisition_bins(self, sequencer_name, bins):
        seq = self.readouts[sequencer_name]
        q1asm = self._loaded_q1asm[sequencer_name]
        if q1asm is None or len(q1asm['acquisitions']) == 0:
            logger.warning(f'No acquisitions for {sequencer_name}')
            return None
        module = self.modules[seq.module_name]
        state = module.pulsar.get_acquisition_state(seq.seq_nr, 1)
        logger.info(f'Acquisition status {sequencer_name} ({module.pulsar.name}:'
                     f'{seq.seq_nr}): {state}')
        return module.pulsar.get_acquisitions(seq.seq_nr)[bins]['acquisition']['bins']

    def get_input_ranges(self, sequencer_name):
        ''' Returns input range for both channels of sequencer.
        Value is in Vpp.
        '''
        seq = self.readouts[sequencer_name]
        module = self.modules[seq.module_name]
        in0_gain = module.pulsar.in0_gain.cache()
        in1_gain = module.pulsar.in1_gain.cache()
        in_range = tuple(
                1.0 * 10**(-dB/20)
                for dB in [in0_gain, in1_gain])
        return in_range


def set_exception_on_overload(enable:bool):
    '''
    Setting set_exception_on_overload to False suppresses Exception
    when the QRM input is overloaded.
    '''
    Q1Instrument._exception_on_overload = enable


def check_instrument_status(instrument, print_status=False):
    sys_state = instrument.get_system_state()
    if sys_state.status != 'OKAY':
        if getattr(instrument, 'is_dummy', False):
            print(f'Status (Dummy) {instrument.name}:', sys_state)
        else:
            raise Exception(f'Module {instrument.name} status not OKAY: {sys_state}')



from abc import abstractmethod, ABC

import numpy as np

from .exceptions import Q1TypeError

class Operand(ABC):

    @property
    @abstractmethod
    def dtype(self):
        pass

    def __add__(self, rhs):
        return Addition(self, rhs)

    def __radd__(self, lhs):
        return Addition(lhs, self)

    def __sub__(self, rhs):
        return Subtraction(self, rhs)

    def __rsub__(self, lhs):
        return Subtraction(lhs, self)

    def __lshift__(self, rhs):
        return Asl(self, rhs)

    def __rshift__(self, rhs):
        return Asr(self, rhs)

    def __and__(self, rhs):
        return BitwiseAnd(self, rhs)

    def __rand__(self, lhs):
        return BitwiseAnd(lhs, self)

    def __or__(self, rhs):
        return BitwiseOr(self, rhs)

    def __ror__(self, lhs):
        return BitwiseOr(lhs, self)

    def __xor__(self, rhs):
        return BitwiseXor(self, rhs)

    def __rxor__(self, lhs):
        return BitwiseXor(lhs, self)

    def __invert__(self):
        return BitwiseNot(self)

    def asfloat(self):
        return CastFloat(self)

def get_dtype(value):
    if isinstance(value, (int, np.integer)):
        return int
    if isinstance(value, (float, np.floating)):
        return float
    if isinstance(value, Operand):
        return value.dtype
    # process label as type int.
    if isinstance(value, str) and value[0] == '@':
        return int
    if value is None:
        return type(None)

    return None

class Expression(Operand, ABC):

    @abstractmethod
    def evaluate(self, generator, destination=None):
        pass

class UnaryExpression(Expression):
    def __init__(self, operator, rhs):
        self.operator = operator
        self.rhs = rhs
        self._dtype = self._get_dtype()

    @property
    def dtype(self):
        return self._dtype

    def evaluate(self, generator, destination=None):
        if destination is None:
            destination = generator.get_temp_reg()

        if isinstance(self.rhs, Expression):
            rhs = self.rhs.evaluate(generator)
        else:
            rhs = self.rhs

        self._evaluate(generator, destination, rhs)

        return destination

    @abstractmethod
    def _get_dtype(self):
        pass

    @abstractmethod
    def _evaluate(self, generator, destination, rhs):
        pass

    def __repr__(self):
        return f'{self.operator}{self.rhs}'



class BinaryExpression(Expression):
    def __init__(self, lhs, operator, rhs):
        self.lhs = lhs
        self.operator = operator
        self.rhs = rhs
        self._dtype = self._get_dtype()

    @property
    def dtype(self):
        return self._dtype

    def evaluate(self, generator, destination=None):
        if destination is None:
            destination = generator.get_temp_reg()
        if isinstance(self.lhs, Expression):
            lhs = self.lhs.evaluate(generator)
        else:
            lhs = self.lhs

        if isinstance(self.rhs, Expression):
            rhs = self.rhs.evaluate(generator)
        else:
            rhs = self.rhs

        self._evaluate(generator, destination, lhs, rhs)

        return destination

    @abstractmethod
    def _get_dtype(self):
        pass

    @abstractmethod
    def _evaluate(self, generator, destination, lhs, rhs):
        pass

    def __repr__(self):
        return f'{self.lhs} {self.operator} {self.rhs}'


class Addition(BinaryExpression):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, '+', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.add(lhs, rhs, destination)

    def _get_dtype(self):
        lhs_dtype = get_dtype(self.lhs)
        rhs_dtype = get_dtype(self.rhs)
        if lhs_dtype != rhs_dtype:
            raise Q1TypeError(f'incompatible data types: {self}, '
                              f'{lhs_dtype.__name__} <> {rhs_dtype.__name__}')
        return lhs_dtype

class Subtraction(BinaryExpression):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, '-', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.sub(lhs, rhs, destination)

    def _get_dtype(self):
        lhs_dtype = get_dtype(self.lhs)
        rhs_dtype = get_dtype(self.rhs)
        if lhs_dtype != rhs_dtype:
            raise Q1TypeError(f'incompatible data types: {self}, '
                              f'{lhs_dtype.__name__} <> {rhs_dtype.__name__}')
        return lhs_dtype

class Lsr(BinaryExpression):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, 'unsigned >>', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.lsr(lhs, rhs, destination)

    def _get_dtype(self):
        rhs_dtype = get_dtype(self.rhs)
        if rhs_dtype != int:
            raise Q1TypeError(f'Shift requires integer number of bits {rhs_dtype.__name__}')
        return get_dtype(self.lhs)

class Asr(BinaryExpression):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, '>>', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.asr(lhs, rhs, destination)

    def _get_dtype(self):
        rhs_dtype = get_dtype(self.rhs)
        if rhs_dtype != int:
            raise Q1TypeError(f'Shift requires integer number of bits {self}')
        return get_dtype(self.lhs)

class Asl(BinaryExpression):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, '<<', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.asl(lhs, rhs, destination)

    def _get_dtype(self):
        rhs_dtype = get_dtype(self.rhs)
        if rhs_dtype != int:
            raise Q1TypeError(f'Shift requires integer number of bits {self}')
        return get_dtype(self.lhs)


class Bitwise(BinaryExpression, ABC):
    def __init__(self, lhs, operator, rhs):
        super().__init__(lhs, operator, rhs)

    def _get_dtype(self):
        lhs_dtype = get_dtype(self.lhs)
        rhs_dtype = get_dtype(self.rhs)
        if rhs_dtype != int or lhs_dtype != int:
            raise Q1TypeError(f'Bitwise operation requires integer values {self}')
        return int

class BitwiseAnd(Bitwise):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, '&', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.bits_and(lhs, rhs, destination)

class BitwiseOr(Bitwise):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, '|', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.bits_or(lhs, rhs, destination)

class BitwiseXor(Bitwise):
    def __init__(self, lhs, rhs):
        super().__init__(lhs, '^', rhs)

    def _evaluate(self, generator, destination, lhs, rhs):
        generator.bits_xor(lhs, rhs, destination)

class BitwiseNot(UnaryExpression, ABC):
    def __init__(self, rhs):
        super().__init__('~', rhs)

    def _get_dtype(self):
        rhs_dtype = get_dtype(self.rhs)
        if rhs_dtype != int:
            raise Q1TypeError(f'Bitwise operation requires integer values {self}')
        return int

    def _evaluate(self, generator, destination, rhs):
        generator.bits_not(rhs, destination)

class CastFloat(UnaryExpression, ABC):
    def __init__(self, rhs):
        super().__init__('float ', rhs)

    def _get_dtype(self):
        rhs_dtype = get_dtype(self.rhs)
        if rhs_dtype != int:
            raise Q1TypeError(f'Float cast requires integer value')
        return float

    def _evaluate(self, generator, destination, rhs):
        if destination != rhs:
            generator.move(rhs, destination)

